# Copyright (c) 2011-2012 Cloudera, Inc. All rights reserved.

import logging
import time

from cmf.monitor import schema
from cmf.monitor.abstract_hdfs_monitor import AbstractHdfsMonitor
from cmf.monitor.base import DebugStatusReport
import cmf.monitor.datanode.json_parser
import cmf.monitor.datanode.json_parser_jmx
from cmf.monitor.json_metrics_utils import ParseError, add_jvm_metrics
from cmf.monitor.proc_metrics_utils import add_proc_metrics
from cmf.monitor.role_update import RoleUpdate

LOG = logging.getLogger('DataNodeMonitor')

_ROLETYPE = 'DATANODE'

# Section and key names in cloudera-monitor.properties
# Since CM 3.7
_SECTION = 'datanode'
_SERVICE_NAME = 'service_name'
_ROLE_NAME = 'role_name'
_HOST = 'host'
_DFS_DATANODE_HTTP_PORT = 'dfs_datanode_http_port'
_DFS_DATA_DIRS = 'dfs_data_dirs'
_LOG_DIR = 'log_dir'
_COLLECT_INTERVAL = 'collect_interval'
# Added in CM 4.0
_SERVICE_VERSION = 'service_version'

class DataNodeMonitor(AbstractHdfsMonitor):

  def __init__(self, daemon, registry):
    AbstractHdfsMonitor.__init__(self, 
                                 daemon,
                                 registry,
                                 _SECTION,
                                 _DFS_DATANODE_HTTP_PORT)
    self._last_conf_path = None
    self._metrics_tuple = (None, None, None)
    self._last_reported_timestamp = None
    self._last_collection_ended = None
    self._last_collection_length = None
    self._role_dir_monitors = [ ]

    # Config data
    self._name = None
    self._service_name = None
    self._service_version = None
    self._dfs_data_dirs = [ ]
    self._log_dir = None

  def __str__(self):
    return "DataNodeMonitor for %s" % (self._name,)

  def _reload_config(self, confparse):
    """Returns True/False for success"""
    try:
      # Extraction
      name = confparse.get(_SECTION, _ROLE_NAME)
      service_name = confparse.get(_SECTION, _SERVICE_NAME)
      service_version = \
          confparse.getint_with_default(_SECTION, _SERVICE_VERSION, 3)
      dfs_data_dirs = confparse.getlist(_SECTION, _DFS_DATA_DIRS)
      host = confparse.get(_SECTION, _HOST)
      web_port = confparse.getint(_SECTION, _DFS_DATANODE_HTTP_PORT)
      log_dir = confparse.get(_SECTION, _LOG_DIR)
    except ValueError, ex:
      self.record_error(ex)
      LOG.error("Value error in %s: %s" % (confparse.path, ex))
      return False

    metrics_url = self._compute_metrics_url(service_version, host, web_port)
    if metrics_url is None:
      return False

    # Assign
    self._name = name
    self._service_name = service_name
    self._service_version = service_version
    self._dfs_data_dirs = dfs_data_dirs
    self._log_dir = log_dir
    self._metrics_url = metrics_url

    # Make monitors for the log directory and the data directories.
    self._role_dir_monitors = [ ]
    self.create_dir_monitor(self._log_dir, self._role_dir_monitors)
    for d in self._dfs_data_dirs:
      self.create_dir_monitor(d, self._role_dir_monitors)

    return True

  def collect(self):
    """Overrides AbstractMonitor. Collect DN metrics. Invoked asynchronously."""
    start = time.time()

    # Reload config if necessary
    confparse = self.conf_parser
    if self._last_conf_path != confparse.path:
      if not self._reload_config(confparse):
        return
      self._last_conf_path = confparse.path

    # Collect directory info
    for mon in self._role_dir_monitors:
      mon.collect()

    # If our pid value is None, it means that the associated process is not
    # running. If this is the case, we skip metric collection.
    if self._pid is None:
      self._metrics_tuple = (None, None, None)
      return

    LOG.debug("Fetching DataNode metrics at '%s'" % (self._metrics_url,))
    self._metrics_tuple = self.collect_metrics_from_url(self._metrics_url)

    self._last_collection_ended = time.time()
    self._last_collection_length = self._last_collection_ended - start

  def parse_metrics(self, json):
    if self._service_version == 3:
      return cmf.monitor.datanode.json_parser.JsonParser(json)
    elif self._service_version == 4:
      return cmf.monitor.datanode.json_parser_jmx.JsonParser(json)
    raise Exception("Unsupported operation.")

  def get_role_update(self):
    """Returns a RoleUpdate to be sent to the firehose"""
    update = RoleUpdate(name=self._name,
                        roletype=_ROLETYPE,
                        service_name=self._service_name)

    if self.add_metrics_from_webserver(self._metrics_tuple,
                                       self._last_reported_timestamp,
                                       update):
      self._last_reported_timestamp = update.timestamp
                                    
    add_proc_metrics(update, self._pid, LOG)

    for mon in self._role_dir_monitors:
      update.add_dir_update(mon.get_role_dir_update())

    return update

  def add_webserver_metrics_to_role_update(self, update, metrics):
    if self._service_version == 3:
      self._add_webserver_metrics_to_role_update_cdh3(update, metrics)
    elif self._service_version == 4:
      self._add_webserver_metrics_to_role_update_cdh4(update, metrics)

  def _add_webserver_metrics_to_role_update_cdh3(self, update, metrics):
    # The metrics page is sometimes empty if the DN happens to
    # be flushing the metrics when we fetch it.
    if len(metrics.datanode) == 0:
      return

    dn = metrics.datanode         # To save some typing

    update.add_metric(schema.DN_BYTES_READ, dn.bytes_read)
    update.add_metric(schema.DN_BYTES_WRITTEN, dn.bytes_written)
    update.add_metric(schema.DN_NUM_LOCAL_READS, dn.reads_from_local_client)
    update.add_metric(schema.DN_NUM_REMOTE_READS, dn.reads_from_remote_client)
    update.add_metric(schema.DN_NUM_LOCAL_WRITES, dn.writes_from_local_client)
    update.add_metric(schema.DN_NUM_REMOTE_WRITES, dn.writes_from_remote_client)
    update.add_metric(schema.DN_NUM_VOLUME_FAILURES, dn.volumeFailures)
    add_jvm_metrics(update, metrics)

  def _add_webserver_metrics_to_role_update_cdh4(self, update, metrics):
    if metrics.datanode is not None:
      dn = metrics.datanode         # To save some typing
      update.add_metric(schema.DN_BYTES_READ, dn.BytesRead)
      update.add_metric(schema.DN_BYTES_WRITTEN, dn.BytesWritten)
      update.add_metric(schema.DN_NUM_LOCAL_READS, dn.ReadsFromLocalClient)
      update.add_metric(schema.DN_NUM_REMOTE_READS, dn.ReadsFromRemoteClient)
      update.add_metric(schema.DN_NUM_LOCAL_WRITES, dn.WritesFromLocalClient)
      update.add_metric(schema.DN_NUM_REMOTE_WRITES, dn.WritesFromRemoteClient)

    if metrics.datanode_dataset_state is not None:
      update.add_metric(schema.DN_NUM_VOLUME_FAILURES,
                        metrics.datanode_dataset_state.NumFailedVolumes)

    add_jvm_metrics(update, metrics, True)

  def get_debug_status_report(self):
    return DebugStatusReport(config_items=self.conf_parser.items(_SECTION),
                             errors=(self.last_error,),
                             last_collection_ended=self._last_collection_ended,
                             last_collection_length=self._last_collection_length)

  @property
  def dfs_data_dirs(self):
    return self._dfs_data_dirs

  def get_subject_type(self):
    return "DATANODE"
