# Copyright (c) 2011 Cloudera, Inc. All rights reserved.

import logging
import time

from cmf.monitor.abstract_mgmt_monitor import AbstractMgmtMonitor
from cmf.monitor.base import DebugStatusReport
from cmf.monitor.role_update import RoleUpdate
from cmf.monitor.service_update import ServiceUpdate

from cmf.monitor.proc_metrics_utils import add_proc_metrics

LOG = logging.getLogger('ReportsManagerMonitor')

_ROLETYPE = "REPORTSMANAGER"

# Section and key names in cloudera-monitor.properties
# Since 4.0
_SECTION = 'reportsmanager'
_SERVICE_NAME = 'service_name'
_ROLE_NAME = 'role_name'
_HOST = 'host'
_COLLECT_INTERVAL = 'collect_interval'
# Added in CM 4.1
_LOG_DIR = 'log_dir'
_SCRATCH_DIR = "scratch_dir"

class ReportsManagerMonitor(AbstractMgmtMonitor):

  def __init__(self, daemon, registry):
    AbstractMgmtMonitor.__init__(self,
                                 daemon,
                                 registry,
                                 _SECTION,
                                 None)
    self._last_conf_path = None
    self._last_collection_ended = None
    self._last_collection_length = None

    # Config data
    self._name = None
    self._service_name = None

  def __str__(self):
    return "ReportsManagerMonitor for %s" % (self._name,)

  def _reload_config(self, confparse):
    """Overrides AbstractMonitor. Returns True/False for success"""
    try:
      service = confparse.get(_SECTION, _SERVICE_NAME)
      name = confparse.get(_SECTION, _ROLE_NAME)
      log_dir = confparse.get_with_default(_SECTION, _LOG_DIR, None)
      scratch_dir = confparse.get_with_default(_SECTION, _SCRATCH_DIR, None)
      host = confparse.get(_SECTION, _HOST)
    except ValueError, ex:
      self.record_error(ex)
      return False

    self._service_name = service
    self._name = name

    # Make a monitor for the log directory.
    self._role_dir_monitors = [ ]
    if log_dir is not None:
      self.create_dir_monitor(log_dir, self._role_dir_monitors)
    if scratch_dir is not None:
      self.create_dir_monitor(scratch_dir, self._role_dir_monitors)

    return True

  def get_role_update(self):
    """Returns a RoleUpdate to be sent to the firehose"""
    update = RoleUpdate(name=self._name,
                        roletype=_ROLETYPE,
                        service_name=self._service_name)

    add_proc_metrics(update, self._pid, LOG)

    for mon in self._role_dir_monitors:
      update.add_dir_update(mon.get_role_dir_update())

    return update

  def get_debug_status_report(self):
    return DebugStatusReport(config_items=self.conf_parser.items(_SECTION),
                             errors=(self.last_error,),
                             last_collection_ended=self._last_collection_ended,
                             last_collection_length=self._last_collection_length)

  def collect(self):
    """Overrides AbstractMonitor. Collect metrics. Invoked asynchronously."""
    start = time.time()

    # Reload config if necessary
    confparse = self.conf_parser
    if self._last_conf_path != confparse.path:
      if not self._reload_config(confparse):
        return
      self._last_conf_path = confparse.path

    # Collect directory info
    for mon in self._role_dir_monitors:
      mon.collect()

    # If our pid value is None, it means that the associated process is not
    # running. If this is the case, we skip metric collection.
    if self._pid is None:
      return

    self._last_collection_ended = time.time()
    self._last_collection_length = self._last_collection_ended - start

  def is_ready_to_report(self):
    if self._pid is None:
      return False
    return True

  def get_subject_type(self):
    return "REPORTSMANAGER"
